local Ctrl = require ( 'ctrl' )

local state = {
	menu = { },
	basic = { },
	options = { },
	rebind = { },
}
state.current = state.menu

local grid

local function setgrid ( data, aspect )
	grid = { }
	local w, h = 0, #data
	for i = 1, h do w = math.max ( w, #data[ i ] ) end
	grid.width, grid.height = w, h

	local sw, sh = love.graphics.getDimensions ( )
	grid.xoffset, grid.yoffset = sw * 0.2, sh * 0.2

	grid.xcellspacing = sw * 0.6 / grid.width
	grid.ycellspacing = grid.xcellspacing * aspect

	if grid.ycellspacing * h > sh * 0.6 then
		grid.ycellspacing = sh * 0.6 / grid.height
		grid.xcellspacing = grid.ycellspacing / aspect
		grid.xoffset = sw * 0.2 + ( sw * 0.6 - grid.xcellspacing * w ) / 2
	else
		grid.yoffset = sh * 0.2 + ( sh * 0.6 - grid.ycellspacing * h ) / 2
	end

	grid.cellpadding = math.max ( 5, math.min ( grid.xcellspacing, grid.ycellspacing ) * 0.05 )

	for y = 1, #data do
		grid[ y ] = { xoffset = ( grid.width - #data[ y ] ) * grid.xcellspacing / 2 }
		for x = 1, #data[ y ] do
			grid[ y ][ x ] = { data[ y ][ x ], 0, 0.1 }
		end
	end
end

local function drawgrid ( offset )
	local ww = grid.xcellspacing - grid.cellpadding * 2
	local hh = grid.ycellspacing - grid.cellpadding * 2
	local fh = love.graphics.getFont ( ):getHeight ( ) / 2
	for y = 1, #grid do
		local yy = grid.yoffset + ( y - 1 ) * grid.ycellspacing + grid.cellpadding
		for x = 1, #grid[ y ] do
			local xx = grid.xoffset + ( x - 1 ) * grid.xcellspacing + grid.cellpadding + ( offset and grid[ y ].xoffset or 0 )

			local p1, p2 = 0, grid[ y ][ x ][ 2 ]
			if p2 < 0 then p1, p2 = 1 + p2, -p2 end
			p1, p2 = math.max ( math.min ( p1, 1 ), 0 ), math.max ( math.min ( p2, 1 ), 0 )

			local r, g, b = 100, 200, 150
			if grid.selected == grid[ y ][ x ] then r, g, b = 0, 255, 0 end

			love.graphics.setColor ( r, g, b, grid[ y ][ x ][ 3 ] * 255 )
			love.graphics.rectangle ( "fill", xx, yy, ww, hh )
			love.graphics.setColor ( r, g, b )
			love.graphics.rectangle ( "line", xx, yy, ww, hh )

			love.graphics.rectangle ( "fill", xx + ww * p1, yy, ww * p2, hh )
			love.graphics.printf ( grid[ y ][ x ][ 1 ], xx, yy + hh * 0.5 - fh, ww, "center" )
			love.graphics.setColor ( 0, 0, 0 )
			love.graphics.setScissor ( xx + ww * p1, yy, ww * p2, hh )
			love.graphics.printf ( grid[ y ][ x ][ 1 ], xx, yy + hh * 0.5 - fh, ww, "center" )
			love.graphics.setScissor ( )
		end
	end
end

local function selectgrid ( x, y, offset )
	local ww = grid.xcellspacing - grid.cellpadding * 2
	local hh = grid.ycellspacing - grid.cellpadding * 2
	for iy = 1, #grid do
		local yy = grid.yoffset + ( iy - 1 ) * grid.ycellspacing + grid.cellpadding
		for ix = 1, #grid[ iy ] do
			local xx = grid.xoffset + ( ix - 1 ) * grid.xcellspacing + grid.cellpadding + ( offset and grid[ iy ].xoffset or 0 )
			if x > xx and x < xx + ww and y > yy and y < yy + hh then
				grid.selected = grid[ iy ][ ix ]
				return
			end
		end
	end
	grid.selected = nil
end

--== MENU ==--
do
	local map = { }
	map[ "Basic Example" ] = "basic"
	map[ "Bind Options" ] = "options"
	map[ "Settings Screen" ] = "rebind"

	function state.menu.load ( )
		setgrid ( { { "Basic Example", "Bind Options", "Settings Screen" } }, 0.75 )
	end

	function state.menu.mousereleased ( x, y, button )
		selectgrid ( x, y, true )
		if grid.selected then
			state.current = state[ map[ grid.selected[ 1 ] ] ]
			state.current.load ( )
		end
	end

	function state.menu.mousemoved ( x, y )
		selectgrid ( x, y, true )
	end

	function state.menu.draw ( )
		drawgrid ( true )
	end
end

--== BASIC EXAMPLE ==--
do
	local ctrl

	function state.basic.load ( )
		setgrid ( {
			{ "gas", "brake" },
			{ "left", "right" },
			{ "nitro" },
		}, 0.3 )

		ctrl = Ctrl ( )

		--don't pay too much attention to mapper function right now, it's just there to make this work nicely
		ctrl:bind ( "gas", { "keyboard", "w" } )
		ctrl:bind ( "gas", { "keyboard", "up" } )
		ctrl:bind ( "gas", { "gamepad", "default", "button", "a" } )
		ctrl:bind ( "gas", { "gamepad", "default", "axis", "triggerright" } )
		ctrl:bind ( "gas", { "gamepad", "default", "axis", "righty" },
			{ mapper = { func = function ( raw ) if raw < 0 then return -raw else return 0 end end } } )

		ctrl:bind ( "brake", { "keyboard", "s" } )
		ctrl:bind ( "brake", { "keyboard", "down" } )
		ctrl:bind ( "brake", { "gamepad", "default", "button", "x" } )
		ctrl:bind ( "brake", { "gamepad", "default", "axis", "triggerleft" } )
		ctrl:bind ( "brake", { "gamepad", "default", "axis", "righty" },
			{ mapper = { func = function ( raw ) if raw > 0 then return raw else return 0 end end } } )

		ctrl:bind ( "left", { "keyboard", "a" } )
		ctrl:bind ( "left", { "keyboard", "left" } )
		ctrl:bind ( "left", { "gamepad", "default", "hat", "left" } )
		ctrl:bind ( "left", { "gamepad", "default", "axis", "leftx" },
			{ mapper = { func = function ( raw ) if raw < 0 then return -raw else return 0 end end } } )

		ctrl:bind ( "right", { "keyboard", "d" } )
		ctrl:bind ( "right", { "keyboard", "right" } )
		ctrl:bind ( "right", { "gamepad", "default", "hat", "right" } )
		ctrl:bind ( "right", { "gamepad", "default", "axis", "leftx" },
			{ mapper = { func = function ( raw ) if raw > 0 then return raw else return 0 end end } } )

		ctrl:bind ( "nitro", { "keyboard", "space" } )
		ctrl:bind ( "nitro", { "keyboard", "kp0" } )
		ctrl:bind ( "nitro", { "gamepad", "default", "button", "y" } )

		function ctrl:inputpressed ( input, value ) end
		function ctrl:inputreleased ( input, value ) end
		function ctrl:inputmoved ( input, value ) end

		ctrl:hookup ( )
	end

	local car = { x = 100, y = 100, a = 0, v = 0 }

	function state.basic.update ( dt )
		grid[ 1 ][ 1 ][ 2 ] = ctrl:getValue ( "gas" )
		grid[ 1 ][ 2 ][ 2 ] =-ctrl:getValue ( "brake" )
		grid[ 2 ][ 1 ][ 2 ] = ctrl:getValue ( "left" )
		grid[ 2 ][ 2 ][ 2 ] =-ctrl:getValue ( "right" )
		grid[ 3 ][ 1 ][ 2 ] = ctrl:getValue ( "nitro" )

		car.a = car.a + ( ctrl:getValue ( "right" ) - ctrl:getValue ( "left" ) ) * dt * math.min ( car.v, 100 ) / 30
		car.v = car.v + ( ctrl:getValue ( "gas" ) - ctrl:getValue ( "brake" ) ) * dt * 200
		car.v = car.v + ctrl:getValue ( "gas" ) * ctrl:getValue ( "nitro" ) * dt * 400
		car.v = car.v - car.v * 0.5 * dt
		car.x = car.x + math.cos ( car.a ) * car.v * dt
		car.y = car.y + math.sin ( car.a ) * car.v * dt
		if car.x < 0 then car.x = love.graphics.getWidth ( )
		elseif car.x > love.graphics.getWidth ( ) then car.x = 0 end
		if car.y < 0 then car.y = love.graphics.getHeight ( )
		elseif car.y > love.graphics.getHeight ( ) then car.y = 0 end
	end

	function state.basic.draw ( )
		drawgrid ( true )

		love.graphics.setColor ( 100, 100, 100 )
		love.graphics.translate ( car.x, car.y )
		love.graphics.rotate ( car.a )
		love.graphics.rectangle ( "line", -10, -5, 20, 10 )
		love.graphics.circle ( "line", -5, 0, 4 )
		love.graphics.origin ( )
	end
end

--== BIND OPTIONS EXAMPLE ==--
do
	local ctrl

	function state.options.load ( )
		setgrid ( {
			{ "button (Y)", "axis (R T)", "button (space)" },
			{ "deadzone (L X)", "remap (L Y)" },
			{ "ramp slow (R X)", "ramp fast(R Y)", "ramp slow (w/s)", "ramp fast (a/d)" },
			{ "smooth slow (A)", "smooth fast (B)", "smooth slow (q)", "smooth fast (e)" },
			{ "pressed", "released", "moved" },
			{ "noise (L X)", "momentum (L Y)", "edge" }
		}, 0.5 )

		ctrl = Ctrl ( )

		ctrl:bind ( "foo", { "gamepad", "default", "button", "y" } )
		ctrl:bind ( "bar", { "gamepad", "default", "axis", "triggerright" } )
		ctrl:bind ( "baz", { "keyboard", "space" } )

		-- gamepad mappers
		ctrl:bind ( "deadzone", { "gamepad", "default", "axis", "leftx" },
			{ mapper = { func = "deadzone", args = { deadzone = 0.35 } } } )
		ctrl:bind ( "remap", { "gamepad", "default", "axis", "lefty" },
			{ mapper = { func = "remap", args = { rawmin = 1, rawmax = -1, mapmin = 0.2, mapmax = 0.8 } } } )

		-- by default, ctrl assigns digital callbacks to naturally digital inputs (buttons)
		-- and analog callbacks for everything else; combined with filtering this may result
		-- in analog-like values trigger digital input callbacks anyway

		-- gamepad filters
		ctrl:bind ( "smooth1", { "gamepad", "default", "button", "a" },
			{ filter = { func = "smooth", args = { speed = 0.75 } } } )
		ctrl:bind ( "smooth2", { "gamepad", "default", "button", "b" },
			{ filter = { func = "smooth", args = { speed = 2.5 } } } )
		ctrl:bind ( "ramp1", { "gamepad", "default", "axis", "rightx" },
			{ filter = { func = "ramp", args = { speed = 0.5, min = -1, max = 1 } } } )
		ctrl:bind ( "ramp2",  { "gamepad", "default", "axis", "righty" },
			{ filter = { func = "ramp", args = { speed = -1.5, min = -1, max = 1 } } } )

		-- keyboard filters
		ctrl:bind ( "smooth3", { "keyboard", "q" },
			{ filter = { func = "smooth", args = { speed = 0.75 } } } )
		ctrl:bind ( "smooth4", { "keyboard", "e" },
			{ filter = { func = "smooth", args = { speed = 2.5 } } } )

		-- assign analog events explicitly
		ctrl:bind ( "ramp3", { "keyboard", "w" },
			{ filter = { func = "ramp", args = { speed = 0.5, min = -1, max = 1 } },
				events = { { trigger = "moved", handler = "inputmoved" } } } )
		ctrl:bind ( "ramp3", { "keyboard", "s" },
			{ filter = { func = "ramp", args = { speed = -0.5, min = -1, max = 1 } },
				events = { { trigger = "moved", handler = "inputmoved" } } } )
		ctrl:bind ( "ramp4",  { "keyboard", "a" },
			{ filter = { func = "ramp", args = { speed = -1.5, min = -1, max = 1 } },
				events = { { trigger = "moved", handler = "inputmoved" } } } )
		ctrl:bind ( "ramp4",  { "keyboard", "d" },
			{ filter = { func = "ramp", args = { speed = 1.5, min = -1, max = 1 } },
				events = { { trigger = "moved", handler = "inputmoved" } } } )

		-- standard events
		function ctrl:inputpressed ( input, value )
			grid[ 5 ][ 1 ][ 1 ] = "pressed\n" .. input
			grid[ 5 ][ 1 ][ 3 ] = 0.7
		end

		function ctrl:inputreleased ( input, value )
			grid[ 5 ][ 2 ][ 1 ] = "released\n" .. input
			grid[ 5 ][ 2 ][ 3 ] = 0.7
		end

		function ctrl:inputmoved ( input, value )
			grid[ 5 ][ 3 ][ 1 ] = "moved\n" .. input
			grid[ 5 ][ 3 ][ 3 ] = 0.7
		end

		-- custom mapper and filter functions
		ctrl:addMapperFunction ( "noise",
			function ( raw, last, args ) return love.math.noise ( raw * ( args.scale or 3 ) ) end )
		ctrl:addFilterFunction ( "momentum",
			function ( dt, raw, last, args )
				args.momentum = ( args.momentum or 0 ) + ( raw - last ) * dt * ( args.speed or 1 )
				args.momentum = args.momentum - args.momentum * dt * ( args.damp or 1 )
				return last + args.momentum end )
		-- custom event trigger
		ctrl:addTriggerFunction ( "edge",
			function ( curr, last ) if curr > last and curr > 0.5 and last <= 0.5 then return true end end )

		-- custom event
		function ctrl:inputedge ( input, value )
			grid[ 6 ][ 3 ][ 1 ] = "edge\n" .. input
			grid[ 6 ][ 3 ][ 3 ] = 0.7
		end

		-- use custom functions and events
		ctrl:bind ( "noise", { "gamepad", "default", "axis", "leftx" },
			{ mapper = { func = "noise", args = { scale = 5 } }, events =
				{ eventedge1 = { trigger = "edge", handler = "inputedge" } } } )
		ctrl:bind ( "momentum", { "gamepad", "default", "axis", "lefty" },
			{ filter = { func = "momentum", args = { speed = 1, damp = 3 } }, events =
				{ eventedge1 = { trigger = "edge", handler = "inputedge" } } } )

		ctrl:hookup ( )
	end

	function state.options.update ( dt )
		grid[ 1 ][ 1 ][ 2 ] = ctrl:getValue ( "foo" )
		grid[ 1 ][ 2 ][ 2 ] = ctrl:getValue ( "bar" )
		grid[ 1 ][ 3 ][ 2 ] = ctrl:getValue ( "baz" )

		grid[ 2 ][ 1 ][ 2 ] = ctrl:getValue ( "deadzone" )
		grid[ 2 ][ 2 ][ 2 ] = ctrl:getValue ( "remap" )

		grid[ 3 ][ 1 ][ 2 ] = ctrl:getValue ( "ramp1" )
		grid[ 3 ][ 2 ][ 2 ] = ctrl:getValue ( "ramp2" )
		grid[ 3 ][ 3 ][ 2 ] = ctrl:getValue ( "ramp3" )
		grid[ 3 ][ 4 ][ 2 ] = ctrl:getValue ( "ramp4" )
		grid[ 4 ][ 1 ][ 2 ] = ctrl:getValue ( "smooth1" )
		grid[ 4 ][ 2 ][ 2 ] = ctrl:getValue ( "smooth2" )
		grid[ 4 ][ 3 ][ 2 ] = ctrl:getValue ( "smooth3" )
		grid[ 4 ][ 4 ][ 2 ] = ctrl:getValue ( "smooth4" )
		grid[ 6 ][ 1 ][ 2 ] = ctrl:getValue ( "noise" )
		grid[ 6 ][ 2 ][ 2 ] = ctrl:getValue ( "momentum" )

		grid[ 5 ][ 1 ][ 3 ] = math.max ( 0.1, grid[ 5 ][ 1 ][ 3 ] - dt )
		grid[ 5 ][ 2 ][ 3 ] = math.max ( 0.1, grid[ 5 ][ 2 ][ 3 ] - dt )
		grid[ 5 ][ 3 ][ 3 ] = math.max ( 0.1, grid[ 5 ][ 3 ][ 3 ] - dt )
		grid[ 6 ][ 3 ][ 3 ] = math.max ( 0.1, grid[ 6 ][ 3 ][ 3 ] - dt )
	end

	function state.options.draw ( )
		drawgrid ( true )
	end
end

--== KEY BINDING SCREEN EXAMPLE ==--
do
	local ctrl

	function state.rebind.load ( )
		setgrid ( {
			{ "gas\nw or A", "brake\ns or X" },
			{ "left\na or LEFTX", "right\nd or LEFTX" },
			{ "nitro\nspace or Y" },
		}, 0.3 )

		ctrl = Ctrl ( )

		ctrl:bind ( "gas", { "keyboard", "w" } )
		ctrl:bind ( "gas", { "gamepad", "default", "button", "a" } )
		ctrl:bind ( "brake", { "keyboard", "s" } )
		ctrl:bind ( "brake", { "gamepad", "default", "button", "x" } )
		ctrl:bind ( "left", { "keyboard", "a" } )
		ctrl:bind ( "left", { "gamepad", "default", "axis", "leftx" },
			{ mapper = { func = function ( raw ) return raw < 0 and -raw or 0 end } } )
		ctrl:bind ( "right", { "keyboard", "d" } )
		ctrl:bind ( "right", { "gamepad", "default", "axis", "leftx" },
			{ mapper = { func = function ( raw ) return raw > 0 and raw or 0 end } } )
		ctrl:bind ( "nitro", { "keyboard", "space" } )
		ctrl:bind ( "nitro", { "gamepad", "default", "button", "y" } )

		function ctrl:inputpressed ( input, value ) end
		function ctrl:inputreleased ( input, value ) end
		function ctrl:inputmoved ( input, value ) end

		ctrl:hookup ( )
	end

	local car = { x = 100, y = 100, a = 0, v = 0 }

	function state.rebind.update ( dt )
		grid[ 1 ][ 1 ][ 2 ] = ctrl:getValue ( "gas" )
		grid[ 1 ][ 2 ][ 2 ] =-ctrl:getValue ( "brake" )
		grid[ 2 ][ 1 ][ 2 ] = ctrl:getValue ( "left" )
		grid[ 2 ][ 2 ][ 2 ] =-ctrl:getValue ( "right" )
		grid[ 3 ][ 1 ][ 2 ] = ctrl:getValue ( "nitro" )

		car.a = car.a + ( ctrl:getValue ( "right" ) - ctrl:getValue ( "left" ) ) * dt * math.min ( car.v, 100 ) / 30
		car.v = car.v + ( ctrl:getValue ( "gas" ) - ctrl:getValue ( "brake" ) ) * dt * 200
		car.v = car.v + ctrl:getValue ( "gas" ) * ctrl:getValue ( "nitro" ) * dt * 400
		car.v = car.v - car.v * 0.5 * dt
		car.x = car.x + math.cos ( car.a ) * car.v * dt
		car.y = car.y + math.sin ( car.a ) * car.v * dt
		if car.x < 0 then car.x = love.graphics.getWidth ( )
		elseif car.x > love.graphics.getWidth ( ) then car.x = 0 end
		if car.y < 0 then car.y = love.graphics.getHeight ( )
		elseif car.y > love.graphics.getHeight ( ) then car.y = 0 end
	end

	function state.rebind.mousereleased( x, y, button )
		selectgrid ( x, y, true )
		if button == 1 and grid.selected then
			local function callback ( input, name )
				for y = 1, #grid do for x = 1, #grid[ y ] do grid[ y ][ x ][ 3 ] = 0.1 end end
				-- cancel if user hits backspace
				if input[ 1 ] == "keyboard" and input[ 2 ] == "backspace" then
					ctrl:grab ( )
					return
				end

				-- make new bindings description string
				local existingbinds, bindstr = ctrl:getBindings ( name )

				-- add newest binding to the list
				existingbinds[ #existingbinds + 1 ] = { input = input }
				for i = 1, #existingbinds do
					local bindname
					if existingbinds[ i ].input[ 1 ] == "keyboard" then
						bindname = existingbinds[ i ].input[ 2 ]
					else
						bindname = string.upper ( existingbinds[ i ].input[ 4 ] )
					end

					if not bindstr then bindstr = bindname else bindstr = bindstr .. " or " .. bindname end
				end

				-- find grid block to change
				local selected
				for y = 1, #grid do for x = 1, #grid[ y ] do if grid[ y ][ x ][ 1 ]:match ( "[^\n.]+" ) == name then selected = grid[ y ][ x ] end end end

				selected[ 1 ] = name .. "\n" .. bindstr
			end
			-- only allow input from keyboard and gamepad
			-- automatically bind to pre-selected input
			local input = grid.selected[ 1 ]:match ( "[^\n.]+" )
			ctrl:grab ( callback, input, { { "keyboard" }, { "gamepad" } } )
			grid.selected[ 3 ] = 0.5
		elseif grid.selected then
			-- cancel any existing grab
			ctrl:grab ( )
			for y = 1, #grid do for x = 1, #grid[ y ] do grid[ y ][ x ][ 3 ] = 0.1 end end

			-- right mouse click = unbind everything
			if button == 2 then
				grid.selected[ 1 ] = grid.selected[ 1 ]:match ( "[^\n.]+" )
				ctrl:unbind ( grid.selected[ 1 ] )
			end
		end
	end

	function state.rebind.mousemoved ( x, y )
		selectgrid ( x, y, true )
	end

	function state.rebind.draw ( )
		drawgrid ( true )

		love.graphics.setColor ( 100, 100, 100 )
		love.graphics.translate ( car.x, car.y )
		love.graphics.rotate ( car.a )
		love.graphics.rectangle ( "line", -10, -5, 20, 10 )
		love.graphics.circle ( "line", -5, 0, 4 )
		love.graphics.origin ( )
	end
end

--== MAIN ==--
local function resetFunctions ( )
	function love.load ( )
		state.current.load ( )
	end

	function love.keypressed ( key, code )
		if code == "escape" then
			if state.current == state.menu then
				love.event.quit ( )
			end
			resetFunctions ( )
			state.current = state.menu
			state.current.load ( )
		end
	end

	function love.mousepressed ( x, y, button )
		if state.current.mousepressed then state.current.mousepressed ( x, y, button ) end
	end

	function love.mousereleased ( x, y, button )
		if state.current.mousereleased then state.current.mousereleased ( x, y, button ) end
	end

	function love.mousemoved ( x, y )
		if state.current.mousemoved then state.current.mousemoved ( x, y ) end
	end

	function love.update ( dt )
		if state.current.update then state.current.update ( dt ) end
	end

	function love.draw ( )
		state.current.draw ( )
	end
end

resetFunctions ( )
